#define INCL_WIN
#include <os2.h>

#include "blitters.h"
#include "dualmode.h"
#include "mglserver_internal.h"

extern struct MGLSC_state mglclient;
extern int view_x1, view_y1, view_x2, view_y2;

static RECTL junkrectl = { 0, 0, 0, 0 };

#define maker565( r, g, b ) (unsigned short) (((unsigned short)(b)>>3) | ((unsigned short)(g)&0x00fc)<<3 | ((unsigned short)(r)&0x00f8)<<8)
#define makergb4( r, g, b ) (unsigned short) (((unsigned char)(b)) | ((unsigned char)(g)<<8) | ((unsigned char)(r)<<16))

#define redcomp32( data32 )   ((data32) & 255)
#define greencomp32( data32 ) ((data32 >> 8) & 255)
#define bluecomp32( data32 )  ((data32 >> 16) & 255)

#define redcomp16( data16 )   ((data16 >> 8) & 248)
#define greencomp16( data16 ) ((data16 >> 3) & 252)
#define bluecomp16( data16 )  ((data16 << 3) & 248)


#define DECLARE_VARS( DESTTYPE, SRCTYPE )                     \
  unsigned DESTTYPE *framebuf;                                \
  unsigned SRCTYPE *srcbuf = mglclient.vidbuffer, *basestart; \
                                                              \
  int i, j;                                                   \
                                                              \
  int srcstartx, srcstopx, deststartx, deststopx;             \
  int srcstarty, srcstopy, deststarty, deststopy;             \
  int destwidth, destheight;                                  \
  int srcwidth, srcheight, sofarx, sofary;                    \
  float accumulator;                                          \
  float multx, multy;


#define GENERIC_CLIPPING_SCALING( SRCMULT )                         \
  multx = (float) (view_x2 - view_x1) / (float) mglclient.winsizex; \
  multy = (float) (view_y2 - view_y1) / (float) mglclient.winsizey; \
                                                                    \
  if ( mglclient.winposx < 0 )                                      \
  {                                                                 \
    srcstartx  = view_x1 + ((-mglclient.winposx) * multx);          \
    deststartx = 0;                                                 \
  } else {                                                          \
    srcstartx  = view_x1;                                           \
    deststartx = mglclient.winposx;                                 \
  }                                                                 \
                                                                    \
  if ( (mglclient.winposx + mglclient.winsizex) >                   \
        mglclient.desktopscan )                                     \
  {                                                                 \
    srcstopx  = view_x1 +                                           \
     ((mglclient.desktopscan - mglclient.winposx) * multx);         \
    deststopx = mglclient.desktopscan;                              \
  } else {                                                          \
    srcstopx  = view_x2;                                            \
    deststopx = mglclient.winposx + mglclient.winsizex;             \
  }                                                                 \
                                                                    \
  if ( mglclient.winposy < 0 )                                      \
  {                                                                 \
    srcstarty  = view_y1 + ((-mglclient.winposy) * multy);          \
    deststarty = 0;                                                 \
  } else {                                                          \
    srcstarty  = view_y1;                                           \
    deststarty = mglclient.winposy;                                 \
  }                                                                 \
                                                                    \
  if ( (mglclient.winposy + mglclient.winsizey) >                   \
        mglclient.desktopheight )                                   \
  {                                                                 \
    srcstopy  = view_y1 +                                           \
     ((mglclient.desktopheight - mglclient.winposy) * multy);       \
    deststopy = mglclient.desktopheight;                            \
  } else {                                                          \
    srcstopy  = view_y2;                                            \
    deststopy = mglclient.winposy + mglclient.winsizey;             \
  }                                                                 \
                                                                    \
  destwidth  = deststopx - deststartx;                              \
  destheight = deststopy - deststarty;                              \
  srcwidth  = srcstopx - srcstartx;                                 \
  srcheight = srcstopy - srcstarty;                                 \
                                                                    \
  srcbuf += (srcstartx + (srcstarty * mglclient.width)) * SRCMULT;  \
                                                                    \
  basestart = srcbuf;

#define BLITTER_LOOPS( DESTMULT, DESTTYPE, SRCMULT )             \
  DiveAcquireFrameBuffer( mglclient.diveinst, &junkrectl );      \
                                                                 \
  sofary = 0;                                                    \
                                                                 \
  for ( i=0; i<destheight && sofary < srcheight; ++i )           \
  {                                                              \
    sofary = (i * srcheight) / destheight;                       \
    srcbuf = basestart + ((sofary * mglclient.width) * SRCMULT); \
                                                                 \
    accumulator = 0;                                             \
    framebuf = (unsigned DESTTYPE *)mglclient.PMframebuffer +    \
     ((((deststarty + i) * mglclient.desktopscan) + deststartx)  \
      * DESTMULT);                                               \
                                                                 \
    sofarx = 0;                                                  \
                                                                 \
    for ( j=0; j<destwidth && sofarx < srcwidth; ++j )           \
    {

#define END_BLITTER_LOOPS( DESTINC, SRCINC )      \
      framebuf += DESTINC;                        \
      accumulator += multx;                       \
      srcbuf += ((int)accumulator) * SRCINC;      \
      sofarx += (int)accumulator;                 \
      accumulator -= (int)accumulator;            \
    }                                             \
  }                                               \
  DiveDeacquireFrameBuffer( mglclient.diveinst );


void DirectBlit8To8( void )
{
  DECLARE_VARS( char, char );

  GENERIC_CLIPPING_SCALING( 1 );

  BLITTER_LOOPS( 1, char, 1 )
   *framebuf = *srcbuf;
  END_BLITTER_LOOPS( 1, 1 )
}

void DirectBlit8To16( void )
{
  MGL_SERVER_COLORS_SET_PACKET *cs = 
   (MGL_SERVER_COLORS_SET_PACKET *) mglclient.shared_packet;

  DECLARE_VARS( short, char );

  GENERIC_CLIPPING_SCALING( 1 );

  BLITTER_LOOPS( 1, short, 1 )
    *framebuf =
      maker565( cs->colors[*srcbuf].red, cs->colors[*srcbuf].green,
       cs->colors[*srcbuf].blue );
  END_BLITTER_LOOPS( 1, 1 )
}

void DirectBlit8To24( void )
{
  MGL_SERVER_COLORS_SET_PACKET *cs = 
   (MGL_SERVER_COLORS_SET_PACKET *) mglclient.shared_packet;

  DECLARE_VARS( char, char );

  GENERIC_CLIPPING_SCALING( 1 );

  BLITTER_LOOPS( 3, char, 1 )
    *framebuf     = cs->colors[*srcbuf].red;
    *(framebuf+1) = cs->colors[*srcbuf].green;
    *(framebuf+2) = cs->colors[*srcbuf].blue;
  END_BLITTER_LOOPS( 3, 1 )
}

void DirectBlit8To32( void )
{
  MGL_SERVER_COLORS_SET_PACKET *cs = 
   (MGL_SERVER_COLORS_SET_PACKET *) mglclient.shared_packet;

  DECLARE_VARS( long, char );

  GENERIC_CLIPPING_SCALING( 1 );

  BLITTER_LOOPS( 1, long, 1 )
    *framebuf =
     makergb4( cs->colors[*srcbuf].red, cs->colors[*srcbuf].green,
      cs->colors[*srcbuf].blue );
  END_BLITTER_LOOPS( 1, 1 )
}

void DirectBlit16To8( void )
{
  // NOT SUPPORTED.  JUST PAINT THE WINDOW WITH COLOR 0.

  DECLARE_VARS( char, short );

  GENERIC_CLIPPING_SCALING( 1 );

  BLITTER_LOOPS( 1, char, 1 )
    *framebuf = 0;
  END_BLITTER_LOOPS( 1, 1 )
}

void DirectBlit16To16( void )
{
  DECLARE_VARS( short, short );

  GENERIC_CLIPPING_SCALING( 1 );

  BLITTER_LOOPS( 1, short, 1 )
    *framebuf = *srcbuf;
  END_BLITTER_LOOPS( 1, 1 )
}

void DirectBlit16To24( void )
{
  DECLARE_VARS( char, short );

  GENERIC_CLIPPING_SCALING( 1 );

  BLITTER_LOOPS( 3, char, 1 )
    *framebuf       = redcomp16( *srcbuf );
    *(framebuf + 1) = greencomp16( *srcbuf );
    *(framebuf + 2) = bluecomp16( *srcbuf );
  END_BLITTER_LOOPS( 3, 1 )
}

void DirectBlit16To32( void )
{
  DECLARE_VARS( long, short );

  GENERIC_CLIPPING_SCALING( 1 );

  BLITTER_LOOPS( 1, long, 1 )
    *framebuf = makergb4( redcomp16( *srcbuf ), greencomp16( *srcbuf ),
     bluecomp16( *srcbuf ) );
  END_BLITTER_LOOPS( 1, 1 )
}

void DirectBlit24To8( void )
{
  // NOT SUPPORTED.  JUST PAINT THE WINDOW WITH COLOR 0.

  DECLARE_VARS( char, char );

  GENERIC_CLIPPING_SCALING( 3 );

  BLITTER_LOOPS( 1, char, 3 )
    *framebuf = 0;
  END_BLITTER_LOOPS( 1, 3 )
}

void DirectBlit24To16( void )
{
  DECLARE_VARS( short, char );

  GENERIC_CLIPPING_SCALING( 3 );

  BLITTER_LOOPS( 1, short, 3 )
    *framebuf =
     maker565( *srcbuf, *(srcbuf + 1), *(srcbuf + 2) );
  END_BLITTER_LOOPS( 1, 3 )
}

void DirectBlit24To24( void )
{
  DECLARE_VARS( char, char );

  GENERIC_CLIPPING_SCALING( 3 );

  BLITTER_LOOPS( 3, char, 3 )
    *framebuf       = *srcbuf;
    *(framebuf + 1) = *(srcbuf + 1);
    *(framebuf + 2) = *(srcbuf + 2);
  END_BLITTER_LOOPS( 3, 3 )
}

void DirectBlit24To32( void )
{
  DECLARE_VARS( long, char );

  GENERIC_CLIPPING_SCALING( 3 );

  BLITTER_LOOPS( 1, long, 3 )
    *framebuf = makergb4( *srcbuf, *(srcbuf+1), *(srcbuf+2) );
  END_BLITTER_LOOPS( 1, 3 )
}

void DirectBlit32To8( void )
{
  // NOT SUPPORTED.  JUST PAINT THE WINDOW WITH COLOR 0.

  DECLARE_VARS( char, long );

  GENERIC_CLIPPING_SCALING( 1 );

  BLITTER_LOOPS( 1, char, 1 )
    *framebuf = 0;
  END_BLITTER_LOOPS( 1, 1 )
}

void DirectBlit32To16( void )
{
  DECLARE_VARS( short, long );

  GENERIC_CLIPPING_SCALING( 1 );

  BLITTER_LOOPS( 1, short, 1 )
    *framebuf =
     maker565( redcomp32(*srcbuf), greencomp32(*srcbuf),
      bluecomp32(*srcbuf) );
  END_BLITTER_LOOPS( 1, 1 )
}

void DirectBlit32To24( void )
{
  DECLARE_VARS( char, long );

  GENERIC_CLIPPING_SCALING( 1 );

  BLITTER_LOOPS( 3, char, 1 )
    *framebuf     = redcomp32( *srcbuf );
    *(framebuf+1) = greencomp32( *srcbuf );
    *(framebuf+2) = bluecomp32( *srcbuf );
  END_BLITTER_LOOPS( 3, 1 )
}

void DirectBlit32To32( void )
{
  DECLARE_VARS( long, long );

  GENERIC_CLIPPING_SCALING( 1 );

  BLITTER_LOOPS( 1, long, 1 )
    *framebuf = *srcbuf;
  END_BLITTER_LOOPS( 1, 1 )
}

